{% extends "qui.html" %}
{% from "qui.html" import css_devel_file, main_script_file %}

{% block style %}
    {% if debug %}
        {% for theme in themes %}
            {{ css_devel_file(theme, "common.css") }}

            {{ css_devel_file(theme, "dashboard/dashboard.css") }}
            {{ css_devel_file(theme, "dashboard/widgets/analog-widget.css") }}
            {{ css_devel_file(theme, "dashboard/widgets/inc-dec-buttons.css") }}
            {{ css_devel_file(theme, "dashboard/widgets/mjpeg-video.css") }}
            {{ css_devel_file(theme, "dashboard/widgets/multi-value-slider.css") }}
            {{ css_devel_file(theme, "dashboard/widgets/on-off-button.css") }}
            {{ css_devel_file(theme, "dashboard/widgets/on-off-indicator.css") }}
            {{ css_devel_file(theme, "dashboard/widgets/push-button.css") }}
            {{ css_devel_file(theme, "dashboard/widgets/radio-buttons.css") }}
            {{ css_devel_file(theme, "dashboard/widgets/slider.css") }}
            {{ css_devel_file(theme, "dashboard/widgets/text-indicator.css") }}

            {{ css_devel_file(theme, "devices/devices.css") }}
        {% endfor %}
    {% endif %}
{% endblock %}

{% block script %}
    {{ main_script_file(
        {"data-slaves-enabled": slaves_enabled|lower, "data-discover-enabled": discover_enabled|lower}
    ) }}
{% endblock %}
